#!/usr/bin/env python3
import getpass
import json
import math
from datetime import datetime, timedelta, date, time
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union, List, cast, Callable
from typing_extensions import Literal
import pandas as pd
from tenacity import retry, stop_after_delay, stop_after_attempt, wait_fixed, retry_if_exception_type
import requests
import teselagen

DEFAULT_HOST_URL: str = "https://platform.teselagen.com"
DEFAULT_API_TOKEN_NAME: str = "x-tg-cli-token"

DEFAULT_MAX_DATAPOINTS: int = 100
TimeUnit = Literal["miliseconds", "seconds", "minutes", "hours", "days"]


def downsample_data(
    dataframe: pd.DataFrame,
    time_column: str,
    max_samples: int = DEFAULT_MAX_DATAPOINTS,
    sampling_period: Optional[int] = None,
    verbose: Optional[bool] = False,
) -> pd.DataFrame:
    """
        This function can down sample any tabular data that contains a time column. Data must be input as a dataframe together with the necessary arguments.
        One can downsample based on a maximum number of sample points expected or by passing a sampling period.

        NOTE: Down sampling removes samples from the dataframe. The criteria used to remove sample points within the sampling period is to keep the first sample
            in each sampling period and remove the rest ('pandas.DataFrame.resample(...).first().dropna()').
        TODO: Add support for other down sampling criteria functions such as 'last', 'average', etc.

        Args:
            dataframe (pd.DataFrame): Input data as a pandas dataframe.
            time_column (str): name of the column that contains the sample time values used as reference for the down sampling.
            sampling_period (int): This input determines the interval by which to downsample the dataframe.
            max_samples (int): Serves as an alternative to 'sampling_period'. If no sampling_period is provided, it will be automatically computed based
                on the maximum number of desired samples.
            verbose (Optional[bool]): If set to 'True', it will print information relevant to the function such the input dataframe's timespan.

        Returns: A down sampled version of the input dataframe.
    """

    _df: pd.DataFrame = dataframe.copy()

    # Create a date time column for decimation with pandas resample. A dummy date with a dummy seconds time unit will be used.
    _df["DateTime"] = _df[f"{time_column}"].apply(
        lambda timevalues: datetime.combine(date.today(), time()) + timedelta(
            seconds=timevalues))

    # Compute the time span in seconds.
    timespan = (cast(datetime, _df["DateTime"].max()) -
                cast(datetime, _df["DateTime"].min())).total_seconds()

    # number of sample points in input dataframe.
    numer_of_datapoints = _df.shape[0]

    # If no sampling period is provided use 'max_samples' to determine the sampling period.
    if sampling_period is None:

        # If the actual number of datapoints in the dataframe or the dataframe's timespan is 'leq'
        # than the max number of datapoints no downsampling is needed.
        is_down_sampling_needed = (numer_of_datapoints >
                                   max_samples) or (timespan > max_samples)

        if is_down_sampling_needed:
            # Compute the sampling period automatically based on the number of datapoints wanted (defaulted to 'DEFAULT_max_samples').
            sampling_period = math.floor(timespan / max_samples)
            _df.set_index("DateTime", inplace=True)
            _df = _df.resample(
                rule=f"{sampling_period}s").first().dropna().reset_index(
                    drop=True)

        # If no down sampling is needed simply return the input dataframe.
        else:
            return dataframe

    # If a sampling period is provided, resample the dataframe by it.
    else:
        _df.set_index("DateTime", inplace=True)
        _df = _df.resample(
            rule=f"{sampling_period}s").first().dropna().reset_index(drop=True)

    if verbose:
        print(f"Data timespan: {timespan}")
        print(f"Sample points left: {_df.shape[0]}")
        print(f"Sample points removed: {dataframe.shape[0] - _df.shape[0]}")

    return _df


def xlsx_parser(
    filepath: Union[str, Path],
    sheet_names: Optional[List[str]] = None,
) -> Dict[str, Union[pd.DataFrame, dict]]:
    '''
        This method takes in a *.xlsx file and converts it to a dictionary with its sheet names as keys
        and the sheet's data as its values. The sheet's data comes as a dataframe.

        Args:
         - filepath (Union[str, Path]): file path to the XLSX file.
         - sheet_names (Optional[List[str]]): Sheet names to process. If None is provided, all sheets will be processed.
    '''
    sheets_data = {}
    reader = pd.ExcelFile(filepath, engine="openpyxl")
    for sheet_name in reader.sheet_names:
        # If sheet_names is provided, skip sheets not in 'sheet_names'
        if isinstance(sheet_names, list) and len(
                sheet_names) > 0 and sheet_name not in sheet_names:
            continue

        print(f"Reading data from sheet: {sheet_name}")
        sheet_df = pd.read_excel(
            str(filepath),
            sheet_name=sheet_name,
            engine="openpyxl",
        )
        sheets_data[sheet_name] = sheet_df

    return sheets_data


def load_from_json(filepath: Path) -> Any:
    """

    Loads a JSON file.

    Args:
        filepath (Path) : Path to the input JSON.

    Returns:
        (Any) : It returns a JSON object.

    """
    absolute_path: Path = filepath.absolute()

    json_obj: Any = json.loads(absolute_path.read_text())

    return json_obj


def get_project_root() -> Path:
    """ Returns project's root folder <absolute/path/to>/lib
    """
    return Path(cast(list, teselagen.__path__)[0]).parent.resolve()


def get_credentials_path() -> Path:
    """ Returns path to where credentials file should be
    """
    return get_project_root() / '.credentials'


def get_test_configuration_path() -> Path:
    """ Returns path to where .test_configuration file should be
    """
    return get_project_root() / '.test_configuration'


## CLIENT UTILS


def get_credentials(
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> Tuple[str, str]:
    """

        It prompts the user for credentials in case username/password aren't provided
        and credentials file wasn't found.

        Args:
            username (Optional[str]) :  A valid username address to authenticate.
                If not provided, it will be prompted.

                Default : None

            password (Optional[str]) : A password to authenticate with. If not
                provided it will be prompted.

                Default : None

        Returns:
            (Tuple[str, str]) : It returns the credentials as a tuple of strings,
                containing the username and password.

                (user, password)
    """
    # Check if crentials are defined on a file
    file_credentials = load_credentials_from_file()
    username = file_credentials[0] if username is None else username
    password = file_credentials[1] if password is None else password
    # If credentials aren't defined, get them from user input
    try:
        username = input(f"Enter username: ") if username is None else username
        password = getpass.getpass(prompt=f"Password for {username}: "
                                  ) if password is None else password
    except IOError as e:
        msg = ("""There was an error with user input. If you are making parallel
               tests, make sure you are avoiding 'input' by adding CREDENTIALS
               file.""")
        raise IOError(msg)
    # End
    return username, password


def load_credentials_from_file(
    path_to_credentials_file: str = None
) -> Tuple[Optional[str], Optional[str]]:
    """Load credentials from json credentials file

    The credentials file should contain a JSON object
    with the following keys (and the values)

    ```
    {
        "username": "user",
        "password": "password"
    }
    ```

    Args:
        path_to_credentials_file (str): Path to the file. If not set it will check for `.credentials` file
            at the folder that holds this method.
    Returns:
        username, password: Username and password strings if info is found in a credentials file, and (None, None)
            if not.
    """
    if path_to_credentials_file is None:
        path_to_credentials_file = str(get_credentials_path())
    if not Path(path_to_credentials_file).is_file():
        return None, None
    credentials: Dict = load_from_json(filepath=Path(path_to_credentials_file))
    return credentials['username'], credentials['password']


def handler(func):
    """

    Decorator to handle the response from a request.

    """

    def wrapper(**kwargs):
        # -> requests.Response
        if "url" not in kwargs.keys():
            message = "url MUST be specified as keyword argument"
            raise Exception(message)

        url: str = kwargs.pop("url")

        try:
            response: requests.Response = func(url, **kwargs)

            if response.ok:
                return response

            elif response.status_code == 400:
                resp = json.loads(response.content)
                message: str = f"{response.reason}: {resp['error']}"
                raise Exception(message)

            elif response.status_code == 401:
                message: str = f"URL : {url} access is unauthorized."
                raise Exception(message)

            elif response.status_code == 404:
                message: str = f"URL : {url} cannot be found."
                raise Exception(message)

            elif response.status_code == 405:
                message: str = f"Method not allowed. URL : {url}"
                raise Exception(message)

            # TODO : Add more exceptions.

            else:
                # reason: str = response.reason
                message: str = f"Got code : {response.status_code}. Reason : {response.reason}"
                raise Exception(message)

        except Exception as e:
            raise

    return wrapper


def parser(func):
    """

    Decorator to parse the response from a request.

    """

    def wrapper(**kwargs) -> Dict[str, Union[str, bool, None]]:

        if "url" not in kwargs.keys():
            message = "url MUST be specified as keyword argument"
            raise Exception(message)

        url: str = kwargs["url"]

        response: requests.Response = func(**kwargs)

        # TODO : Should we get/return JSON Serializables values ?
        # status 204 has no content.
        if response.status_code == 204:
            print("Deletion successful.")
            return {}

        response_as_json: Dict[str, Union[str, bool, None]] = {
            "url": url,
            "status": response.ok,
            "content": response.content.decode() if response.ok else None
        }

        return response_as_json

    return wrapper


def requires_login(func):
    """ Decorator to perform login beforehand, if necessary

    Add this decorator to any function from Client or a children
    that requires to be logged in.
    """

    def wrapper(self, *args, **kwargs):
        if self.auth_token is None:
            self.login()
            if self.auth_token is None:
                raise Exception(
                    "Could not access API, access token missing. Please use the 'login' function to obtain access."
                )
        return func(self, *args, **kwargs)

    return wrapper


@parser
@handler
def get(url: str, params: Dict[str, Any] = None, **kwargs):
    """

    Same arguments and behavior as requests.get but handles exceptions and
    returns a dictionary instead of a requests.Response.

    NOTE : url key MUST be passed in arguments.

    Returns:
        (Dict[str, Union[str, bool, None]]) : It returns a dictionary with the
            following keys and value types:

            {   "url" : str,
                "status" : bool,
                "content" : Optional[str, None]
            }

    Raises:

        (Exception) : It raises an exception if something goes wrong.

    """
    response: requests.Response = requests.get(url, params=params, **kwargs)
    return response


@parser
@handler
def post(url: str, **kwargs) -> requests.Response:
    """

    Same as requests.post but handles exceptions and returns a dictionary
    instead of a requests.Response.

    NOTE : url key MUST be passed in arguments.

    Example :

        url = "https://www.some_url.com/"
        response = post(url=url)

    Wrong usage:

        url = "https://www.some_url.com/"
        response = post(url)

    Returns:

        (Dict[str, Union[str, bool, None]]) : It returns a dictionary with the
            following keys and value types:

            {   "url" : str,
                "status" : bool,
                "content" : Optional[str, None]
            }

    Raises:

        (Exception) : It raises an exception if something goes wrong.

    """
    response: requests.Response = requests.post(url, **kwargs)
    return response


@parser
@handler
def delete(url: str, **kwargs) -> requests.Response:
    """
    Same as requests.delete but handles exceptions and returns a dictionary
    instead of a requests.Response.

    """
    response: requests.Response = requests.delete(url, **kwargs)
    return response


@parser
@handler
def put(url: str, **kwargs):
    response: requests.Response = requests.put(url, **kwargs, timeout=None)
    return response


def download_file(url: str, local_filename: str = None, **kwargs) -> str:
    """ Downloads a file from the specified url
    """
    if local_filename is None:
        local_filename = url.split('/')[-1]
    # NOTE the stream=True parameter below
    chunk_size = None
    with requests.get(url, stream=True, **kwargs) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                if chunk:
                    f.write(chunk)
    return local_filename

def wait_for_status(
    method: Callable, 
    validate: Optional[Callable]=None,
    fixed_wait_time: int = 5,
    timeout: int = 300,
    **method_kwargs)->Any:
    """Tries to run *method* (and run also a validation of its output) until no AssertionError is raised

    Arguments are described below. More keyword arguments can be given for `method`.

    Args:
        method (Callable): An unreliable method (or a status query). The method will be executed
            while: (it raises an AssetionError or the `validation` function outputs `False`) and
            none of the ending conditions is satisfied (look at int arguments)

        validate (Optional[Callable], optional): A callable that validates the output of *method*. 
            It must receives the output of `method`  as argument and returns `True` if it is ok 
            and `False` if it is invalid. Defaults to None, meaning no validation will be executed.

        fixed_wait_time (int, optional): Time (in seconds) to wait between attempts. Defaults to 5.
        
        timeout (int, optional): Time (in seconds) after which no more attempts are made. Defaults 
            to 300 (5 minutes).

    Returns:
        [Any]: The method's output
    """
    @retry(wait=wait_fixed(fixed_wait_time),stop=stop_after_delay(timeout), retry=retry_if_exception_type(AssertionError))
    def _wait_for_status(method: Callable, validate: Optional[Callable]=None, **method_kwargs)->Any:
        """Runs the method and apply validation"""
        try:
            result = method(**method_kwargs)
            if validate is not None:
                assert validate(result), f"Validation failed. Result is {result}"
        except Exception as ex:
            if not isinstance(ex, AssertionError):
                print(f"An unexpected error was detected, method result was: {result}")
            raise ex
        return result

    return _wait_for_status(method=method, validate=validate, **method_kwargs)



#!/usr/bin/env python3
import functools
from typing import Dict, Any, Callable, Optional
from pathlib import Path
import os 


import pytest
import pandas as pd
from tenacity import retry, stop_after_delay, stop_after_attempt, wait_fixed

from teselagen.api import TeselaGenClient
from teselagen.utils.utils import wait_for_status

def delete_file(file_name, client_with_lab, ):
    # Get file id 
    files = client_with_lab.test.get_files_info()
    fileterd_files = [file_i for file_i in files if file_i["name"]==file_name]
    client_with_lab.test.delete_file(file_id=fileterd_files[-1]["id"])

@pytest.fixture(scope="module")
def temp_dir(tmp_path_factory)->Path:
    """This works similar to pytest's testdir but with "module" scope"""
    return tmp_path_factory.mktemp("data")

@pytest.fixture(scope="module")
def client_with_lab(api_token_name, host_url, expiration_time)->TeselaGenClient:
    """ Defines a login and lab selection workflow with a "module" scope"""
    client = TeselaGenClient(api_token_name=api_token_name,
                             host_url=host_url,
                             module_name='test')
    client.login(
        expiration_time=expiration_time,
    )
    client.select_laboratory(lab_name="The Test Lab")
    return client

@pytest.fixture(scope="module")
def wild_type_experiment(client_with_lab: TeselaGenClient) -> Dict[str, Any]:
    """Creates an experiment for "Wild Type" data and destroys it when finished"""
    experiment_name="Test multiomics data for WT Strain"
    experiment = client_with_lab.test.create_experiment(experiment_name=experiment_name)
    yield experiment
    client_with_lab.test.delete_experiment(experiment['id'])

@pytest.fixture(scope="module")
def bio_engineered_experiment(client_with_lab: TeselaGenClient) -> Dict[str, Any]:
    """Creates an experiment for "Bio Engineered" data and destroys it when finished"""
    experiment_name="Test multiomics data for BE Strain"
    experiment = client_with_lab.test.create_experiment(experiment_name=experiment_name)
    yield experiment
    client_with_lab.test.delete_experiment(experiment['id'])

@pytest.fixture(scope="module")
def test_data()->Dict[str, pd.DataFrame]:
    """ Loads all required data for this test module"""
    dir_path = Path(os.path.dirname(os.path.realpath(__file__))) / __name__.split(".")[-1]
    return {
        "EDD_experiment_description_file_WT": pd.read_csv(dir_path / "EDD_experiment_description_file_WT.csv"),
        "EDD_experiment_description_file_BE_designs": pd.read_csv(dir_path / "EDD_experiment_description_file_BE_designs.csv"),
        "EDD_OD_WT": pd.read_csv(dir_path / "EDD_OD_WT.csv"),
        "EDD_external_metabolites_WT": pd.read_csv(dir_path / "EDD_external_metabolites_WT.csv"),
        "EDD_transcriptomics_WTSM": pd.read_csv(dir_path / "EDD_transcriptomics_WTSM.csv"), 
    }

@pytest.fixture(scope="module")
def metadata(client_with_lab: TeselaGenClient, test_data: Dict[str, Any],)->Dict[str, Any]:
    """Builds metadata

    Args:
        client_with_lab (TeselaGenClient): Logged client with a laboratory selected
        test_data (Dict[str, Any]): Test data from files

    Returns:
        Dict[str, Any]: Metadata maps from name to ID
    """
    _metadata = {}
    # Descriptor types
    # Here we are going to create the necessary Descriptor Types 
    # that are going to be used to map the different Strains' charactetristics described in
    # the experiment description files.
    # The first column name is omitted, since it's the 'Line Name' which is not a descriptor but the Strain itself.
    descriptorTypeNames = test_data["EDD_experiment_description_file_WT"].columns.values.tolist()[1:]
    # Here we construct the 'descriptorTypes' metadata records.
    # Also, we strip any leading or trailing spaces in the file header names.
    descriptorTypes = [{"name": descriptorTypeName.strip()} for descriptorTypeName in descriptorTypeNames]
    result = client_with_lab.test.create_metadata(metadataType="descriptorType", metadataRecord=descriptorTypes)
    # After creating the descriptor types, we are going to construct a mapper dictionary
    # that we will use to know the metadata descriptorType record IDs from their names.
    _metadata['descriptor_types'] = {x['name']: x['id'] for x in result}

    # Measurement targets
    # We simply construct a JSON with the 'name' key as below.
    measurementTarget = { "name": "Optical Density" }
    result = client_with_lab.test.create_metadata(metadataType="measurementTarget", metadataRecord=measurementTarget)
    # Again, we here construct this auxiliary mapper dictionary
    # that we will use to know the metadata measurementTarget record ID from its name.
    _metadata['measurement_targets'] = {result[0]['name']: result[0]['id']}

    # Assay subbject class
    # We simply construct a JSON with the 'name' key as below.
    assaySubjectClass = { "name": "Strain" }
    result = client_with_lab.test.create_metadata(metadataType="assaySubjectClass", metadataRecord=assaySubjectClass)
    # Again, we here construct this auxiliary mapper dictionary: 'assaySubjectClassNameToId',
    # that we will use to know the metadata assaySubjectClass record ID from its name.
    _metadata["assay_subject_class"] = {result[0]['name']: result[0]['id']}
    
    # Reference dimension
    # Here we list all the currently available reference dimensions in TEST
    # And see there's already a reference dimension called 'Elapsed Time', which we'll use later on.
    # pprint(client.test.get_metadata(metadataType="referenceDimension"))
    # We are going to store this 'Elapsed Time' ID into a variable to use later.
    _metadata['reference_dimension'] = {'Elapsed Time': '1'}

    # Units
    # First we are going to create this 'dummy' dimensionless unitDimension metadata record.
    result=client_with_lab.test.create_metadata(metadataType="unitDimension", metadataRecord={"name":"dimensionless"})
    unitDimensionId = result[0]['id']
    # Then we are going to create this 'dummy' dimensionless unitScale metadata record.
    result=client_with_lab.test.create_metadata(metadataType="unitScale", metadataRecord={"name":"dimensionless", "unitDimensionId": unitDimensionId})
    unitScales = client_with_lab.test.get_metadata(metadataType="unitScale")
    # Here we just construct an auxiliary mapper dictionary that that we will use 
    # to know the metadata unitScale record ID from its name.
    unitScalesNameToId = {unitScale['name']: unitScale['id'] for unitScale in unitScales}
    # The next units are used by the metabolomics, transcriptomics and proteomics dataset.
    # And these three units are of type Concentration, so we'll add the to the 'Metric Concentration' unit scale.
    # The fourth and last unit called 'n/a', will be used to import the Optical Density data.
    result = client_with_lab.test.create_metadata(metadataType="unit", metadataRecord=[
        {"name":"mM", "unitScaleId": unitScalesNameToId['Metric Concentration']},
        {"name":"FPKM", "unitScaleId": unitScalesNameToId['Metric Concentration']}, 
        {"name":"proteins/cell", "unitScaleId": unitScalesNameToId['Metric Concentration']},
        # we create here the 'n/a' unit with dimensionless (or dummy) scale.
        {"name":"n/a", "unitScaleId": unitScalesNameToId['dimensionless']},
    ])
    _metadata['unit_scales'] = unitScalesNameToId

    return _metadata


@pytest.fixture(scope="module")
def experiment_description_mapper(test_data, metadata):
    """Builds a mapper for experiment descriptions"""
    # This will be our mapper JSON that we are going to construct in a way that we map the file columns accordingly.
    # The mapper JSON is an array of objects. These objects are "structured" header JSON objects.
    # These structured headers include the column's 'name', plus 2 other properties: "class" and "subClass" information.
    # The 'class' property indicates which is the column's metadata class/type, while the "subClass" or "subClassId" 
    # indicates the metadata record ID of such "class".
    _experiment_description_mapper = list()
    for column_name in test_data["EDD_experiment_description_file_WT"].columns.values.tolist():
        if (column_name == "Line Name"):
            structured_header = {
                "name": column_name.strip(),
                "class": "assaySubjectClass",
                "subClassId": metadata["assay_subject_class"]['Strain']
            }
        else:
            structured_header = {
                "name": column_name.strip(),
                "class": "descriptorType",
                "subClassId": metadata["descriptor_types"][column_name.strip()]
            }
        _experiment_description_mapper.append(structured_header)
    return _experiment_description_mapper

@pytest.fixture(scope="module")
def experiment_description_upload(test_data, temp_dir, client_with_lab, experiment_description_mapper):
    """ Uploads data from WT and BE experiments descriptions"""
    # We now have our mapper JSON that describes/maps each column in the file.
    # Now we upload the data
    exp_description_data_names = ["EDD_experiment_description_file_WT", "EDD_experiment_description_file_BE_designs"]
    responses = {}
    for exp_description_data_name in exp_description_data_names:
        # Write data to file
        description_path = temp_dir / f"{exp_description_data_name}.csv"
        test_data[exp_description_data_name].to_csv(description_path, index=False)
        # Send
        responses[exp_description_data_name] = client_with_lab.test.import_assay_subject_descriptors(
            filepath=description_path,
            mapper=experiment_description_mapper,
        )
    # Wait until upload and processing is finished
    for exp_description_data_name in exp_description_data_names:
        _ = wait_for_status(
            method=client_with_lab.test.get_assay_subjects_descriptor_import_status, 
            validate=lambda x: x["content"]["status"]["code"] == "FINISHED", 
            importId=responses[exp_description_data_name]['importId'])

    return responses

@pytest.fixture(scope="module")
def optical_density_upload(metadata, test_data, temp_dir, client_with_lab, wild_type_experiment):
    """ Uploads data from OD experiments"""
    # Prepare mapper
    wt_od_mapper = [
        {
            "name": "Line Name",
            "class": "assaySubjectClass",
            "subClass": metadata["assay_subject_class"]["Strain"]
        },
        {
            "name": "Time",
            "class": "referenceDimension",
            # ID of the referenceDimension metadata record.
            "subClass": metadata["reference_dimension"]['Elapsed Time']
        },
        {
            "name": "Value",
            "class": "measurementTarget",
            # ID of the measurementTarget metadata record.
            "subClass": metadata["measurement_targets"]["Optical Density"]
        },
        {
            "name": "Units",
            "class": "unit",
            # ID of the measurementTarget metadata record.
            # This is in order to assign this "Unit" column to the Value column measurements.
            "subClass": metadata["measurement_targets"]["Optical Density"]
        },
        {
            "name": "time units",
            "class": "d-unit",
            # ID of the referenceDimension metadata record.
            # This is in order to assign this "Unit" column to the Time column measurements.
            "subClass": metadata["reference_dimension"]['Elapsed Time']
        }
    ]

    # Prepare data
    wt_od_df = test_data["EDD_OD_WT"].copy()
    # Adds a "unit" column for Time
    wt_od_df["time units"] = "hrs"
    # Updates the 'Units' column to have the dummy 'n/a' unit created above.
    wt_od_df["Units"] = "n/a"
    # Drops the 'Measurement Type' Columns as it provides no useful information.
    wt_od_df.drop(["Measurement Type"], axis=1, inplace=True)
    # Now we are ready to save this updated dataframe into a new CSV file and upload it into TEST experiment scope.
    new_od_filepath = temp_dir / "TEST_OD_WT.csv"
    wt_od_df.to_csv(new_od_filepath, index=False)

    # Upload data
    # Now we choose to put the assay results into an assay identified by the assay_name variable.
    response = client_with_lab.test.import_assay_results(
        filepath=new_od_filepath,
        assay_name="Wild Type Optical Density",
        experiment_id=wild_type_experiment['id'],
        mapper=wt_od_mapper,
    )

    # Wait until process is finished
    _ = wait_for_status(
        method=client_with_lab.test.get_assay_results_import_status, 
        validate=lambda x: x["content"]["status"]["code"] == "FINISHED", 
        importId=response['importId'])

    yield response

    delete_file(file_name=new_od_filepath.name, client_with_lab=client_with_lab, )

@pytest.fixture(scope="module")
def multiomics_mapper(metadata):
    """ Builds mapper for several multiomics datasets"""
    # We need to construct the multiomic files' structured headers for the mapper JSON object.
    # Here, since the measurement targets are going to be created from the files' "Measurement Type" column values,
    # ee do not specify a subClassId in the structured header of class=measurementTarget.
    return [
        # This first element of the array corresponds to the structured header of the files's "Line Name" column.
        # The four multiomic files have this column and corresponds to the assay subject column of class "Strain".
        {
            "name": "Line Name",
            "class": "assaySubjectClass",
            "subClass": metadata["assay_subject_class"]["Strain"]
        },
        # All four multiomic files have a "Measurement Type" column. Which contains the measurement target values for
        # the 'measurementTarget' metadata class.
        {
            "name": "Measurement Type",
            "class": "measurementTarget",
        },
        # All four multiomic files have a "Time" column. Which represents the reference dimension class.
        {
            "name": "Time",
            "class": "referenceDimension",
            # ID of the referenceDimension metadata record.
            "subClass": metadata["reference_dimension"]["Elapsed Time"]
        },
        # All four multiomic files have a "Value" column. Which contains the measurement values for each
        # measurementTarget metadata record.
        {
            "name": "Value",
            "class": "measurementValue",
        },
        # All four multiomic files have a "Units" column. Which contains the unit for the measurement values for each
        # measurementTarget metadata record.
        {
            "name": "Units",
            "class": "unit",
        },
        # All four multiomic files have a "time units" column. Which contains the unit for the Time reference dimension.
        {
            "name": "time units",
            "class": "d-unit",
            # ID of the referenceDimension metadata record.
            # This is in order to assign this "Unit" column to the Time column measurements.
            "subClass": metadata["reference_dimension"]["Elapsed Time"]
        }
    ]

@pytest.fixture(scope="module")
def upload_external_metabolites( temp_dir, test_data, client_with_lab, wild_type_experiment, multiomics_mapper):
    """Uploads externa metabolites data from files"""
    wt_ext_metabolites_df = test_data["EDD_external_metabolites_WT"].copy()
    # Adds a "unit" column for Time
    client_with_lab.test.get_metadata(metadataType="unit")
    wt_ext_metabolites_df["time units"] = "hrs"
    # Now we are ready to save this updated dataframe into a new CSV file and upload it into TEST experiment scope.
    new_wt_ext_metabolites_filepath = temp_dir/ "TEST_external_metabolites_WT.csv"
    wt_ext_metabolites_df.to_csv(new_wt_ext_metabolites_filepath, index=False)

    # Now we choose to put the assay results into an assay identified by the assay_name variable.
    response = client_with_lab.test.import_assay_results(
        filepath=new_wt_ext_metabolites_filepath, 
        assay_name="Wild Type External Metabolites",
        experiment_id=wild_type_experiment["id"],
        mapper=multiomics_mapper,
    )

    # Wait until process is finished
    _ = wait_for_status(
        method=client_with_lab.test.get_assay_results_import_status, 
        validate=lambda x: x["content"]["status"]["code"] == "FINISHED", 
        importId=response['importId'])

    yield response

    delete_file(file_name=new_wt_ext_metabolites_filepath.name, client_with_lab=client_with_lab, )

@pytest.fixture(scope="module")
def upload_transcriptomics( temp_dir, test_data, client_with_lab, wild_type_experiment, multiomics_mapper):
    """Uploads transcriptomics data from file"""
    wt_transcriptomics_df = test_data["EDD_transcriptomics_WTSM"].copy()
    # Adds a "unit" column for Time
    wt_transcriptomics_df["time units"] = "hrs"
    # Now we are ready to save this updated dataframe into a new CSV file and upload it into TEST experiment scope.
    new_wt_transcriptomics_filepath = temp_dir / "TEST_transcriptomics_WTSM.csv"
    wt_transcriptomics_df.to_csv(new_wt_transcriptomics_filepath, index=False)
    # Now we choose to put the assay results into an assay identified by the assay_name variable.
    response = client_with_lab.test.import_assay_results(
        filepath=new_wt_transcriptomics_filepath, 
        assay_name="Wild Type Transcriptomics",
        experiment_id=wild_type_experiment['id'],
        mapper=multiomics_mapper
    )

    # Wait until process is finished
    _ = wait_for_status(
        method=client_with_lab.test.get_assay_results_import_status, 
        validate=lambda x: x["content"]["status"]["code"] == "FINISHED", 
        importId=response['importId'])

    yield response

    delete_file(file_name=new_wt_transcriptomics_filepath.name, client_with_lab=client_with_lab, )
    

class TestTESTClientMultiomicsData():
    def test_experiment_description_upload(self, experiment_description_upload):
        """Tests data upload methods works ok"""
        # Some assertions
        for _, response in experiment_description_upload.items():
            assert 'importId' in response
            assert 'message' in response

    def test_optical_density_upload(self, optical_density_upload, client_with_lab):
        """Tests data upload methods works ok"""
        # Some assertions
        assert 'importId' in optical_density_upload
        assert 'message' in optical_density_upload
        filtered_assays = [assay for assay in client_with_lab.test.get_assays() if assay["name"]=="Wild Type Optical Density"]
        assert len(filtered_assays)==1, "Expecting just one assay for this assertion"


    def test_upload_external_metabolites(self, upload_external_metabolites, client_with_lab):
        """Tests data upload methods works ok"""
        # Some assertions
        assert 'importId' in upload_external_metabolites
        assert 'message' in upload_external_metabolites
        filtered_assays = [assay for assay in client_with_lab.test.get_assays() if assay["name"]=="Wild Type External Metabolites"]
        assert len(filtered_assays)==1, "Expecting just one assay for this assertion"

    def test_upload_transcriptomics(self, upload_transcriptomics, client_with_lab):
        """Tests data upload methods works ok"""
        # Some assertions
        assert 'importId' in upload_transcriptomics
        assert 'message' in upload_transcriptomics
        filtered_assays = [assay for assay in client_with_lab.test.get_assays() if assay["name"]=="Wild Type Transcriptomics"]
        assert len(filtered_assays)==1, "Expecting just one assay for this assertion"

    def test_download_data(self, upload_transcriptomics, client_with_lab, test_data):
        """Check mapped data is downloaded ok"""
        filtered_assays = [assay for assay in client_with_lab.test.get_assays() if assay["name"]=="Wild Type Transcriptomics"]
        
        # First we download data without subject data
        results_without_subject_data=client_with_lab.test.get_assay_results(
            assay_id=filtered_assays[0]["id"], 
            as_dataframe=True, 
            with_subject_data=False)
        assert len(results_without_subject_data[0]["data"])==9, "Wrong number of output rows"
        assert len(results_without_subject_data[0]["data"].columns)==12, "Wrong number of output columns"

        # Now we download data without subject data
        results_with_subject_data=client_with_lab.test.get_assay_results(
            assay_id=filtered_assays[0]["id"], 
            as_dataframe=True, 
            with_subject_data=True)
        assert len(results_with_subject_data[0]["data"])==9, "Wrong number of output rows"
        assert len(results_with_subject_data[0]["data"].columns)==23, "Wrong number of output columns"

    def test_download_file(self, optical_density_upload, client_with_lab, test_data):
        """Check files download"""
        file_name = "TEST_OD_WT.csv"
        files = client_with_lab.test.get_files_info()
        fileterd_files = [file_i for file_i in files if file_i["name"]==file_name]
        #assert len(fileterd_files)==1, "Expecting just one file for this assertion"
        downloaded = pd.read_csv(client_with_lab.test.download_file(file_id=fileterd_files[0]["id"]))
        assert downloaded.shape == (10, 5), "Wrong shape"
        


#!/usr/bin/env python3
import pytest
import time
from teselagen.utils.utils import wait_for_status
from tenacity import RetryError


class TestUtils():
    def test_wait_for_status(self):
        """Tests wait_for_status happy and sad (timeout) paths with a dummy request"""
        fixed_wait_time = 0.2
        counter_limit = 5
        validation_method = lambda x: x=="OK"
        
        def a_status_request(count_limit: int):
            time.sleep(0.2)
            out = "NOT OK"
            nonlocal counter
            counter += 1
            if counter >= count_limit:
                out = "OK"
            return out

        # Happy path
        counter = 0
        result = wait_for_status(
            method=a_status_request, 
            validate=validation_method, 
            fixed_wait_time=fixed_wait_time, 
            timeout=5, # timeout >= (fixed_wait_time + request_time) * count_limit = (0.2 + 0.2) * 5 = 2.0
            count_limit=counter_limit)
        assert result == "OK"

        # Timeout fail (timeout argument is reduced)
        counter=0
        with pytest.raises(RetryError):# checks exception is raised
            result = wait_for_status(
                method=a_status_request, 
                validate=validation_method, 
                fixed_wait_time=fixed_wait_time, 
                timeout=1, 
                count_limit=counter_limit)
    
    

        


#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT

from io import StringIO
import json
from os.path import join
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal, TypedDict
import warnings

import pandas as pd
from teselagen.utils import delete
from teselagen.utils import get
from teselagen.utils import post
from teselagen.utils import put

# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument


class IAssayResults(TypedDict):
    assayId: str
    fileId: str
    data: Union[pd.DataFrame, List[Dict[str, Any]]]


DEFAULT_PAGE_SIZE: Literal[200] = 200
IMPORTED_FILE_STATUSES: List[str] = ["FINISHED", "FINISHED-DISCARDED"]


class TESTClient():

    def __init__(self, teselagen_client: Any):
        module_name: str = "test"

        self.host_url = teselagen_client.host_url
        self.headers = teselagen_client.headers
        # Here we define the Base CLI URL.
        api_url_base: str = f"{self.host_url}/{module_name}/cli-api"
        # Here we define the client endpoints
        # Example :
        #    self.some_endpoint_url: str = f"{api_url_base}/some_endpoint"

        # Assay Subjects
        self.get_assay_subjects_url: str = f"{api_url_base}/assay-subjects"
        self.get_assay_subject_url: str = join(api_url_base,
                                               "assay-subjects") + "/{}"
        self.create_assay_subjects_url: str = f"{api_url_base}/assay-subjects"
        self.delete_assay_subject_url: str = join(api_url_base,
                                                  "assay-subjects") + "/{}"
        self.put_assay_subject_descriptors_url: str = f"{api_url_base}/assay-subjects/descriptors"

        # Experiments
        self.get_experiments_url: str = f"{api_url_base}/experiments"
        self.create_experiment_url: str = f"{api_url_base}/experiments"
        self.delete_experiment_url: str = join(api_url_base,
                                               "experiments") + "/{}"

        # Assays
        self.get_assays_url: str = f"{api_url_base}/assays"
        self.get_assays_by_experiment_url: str = join(
            api_url_base, "experiments") + "/{}/assays"

        self.create_assay_url: str = join(api_url_base,
                                          "experiments") + "/{}/assays"
        self.delete_assay_url: str = join(api_url_base, "assays") + "/{}"

        self.post_assay_results_import_url: str = f"{api_url_base}/assays/results/import"
        self.get_assay_results_import_url: str = join(
            api_url_base, "assays") + "/results/import/{}"

        self.assay_results_url: str = join(api_url_base,
                                           "assays") + "/{}/results"

        # Files
        self.get_files_info_url: str = f"{api_url_base}/files"
        self.get_file_data_url: str = join(api_url_base, "files") + "/{}"
        self.delete_file_url: str = join(api_url_base, "files") + "/{}"
        self.upload_file_url: str = join(api_url_base, "files")
        self.upload_file_into_assay_url: str = join(api_url_base,
                                                    "assays") + "/{}/files"
        self.upload_file_into_experiment_url: str = join(
            api_url_base, "experiments") + "/{}/files"

        # Metadata
        self.get_metadata_url: str = join(api_url_base, "metadata") + "/{}"
        self.create_metadata_url: str = join(api_url_base, "metadata")
        self.delete_metadata_url: str = join(api_url_base,
                                             "metadata") + "/{}/{}"

    # Assay Subject Endpoints
    def create_assay_subject(self, name: str, assaySubjectClassId: int):
        body = [{"name": name, "assaySubjectClassId": str(assaySubjectClassId)}]

        response: Dict[str, Any] = post(url=self.create_assay_subjects_url,
                                        headers=self.headers,
                                        json=body)

        response["content"] = json.loads(response["content"])

        return response["content"]

    def get_assay_subjects(self,
                           assay_subject_ids: Optional[Union[int,
                                                             List[int]]] = None,
                           summarized: bool = True):
        """
        This function fetches one or many assay subject records from TEST. It receives either an integer ID or a list of integer IDs
        through the 'assay_subject_ids' argument, which correspond to the the assay subject IDs.

        Args:
            assay_subject_ids(Optional[Union[int, List[int]]]): Either an integer, a list of integers or None. When integers are passed,
                these are treated as the assay subject IDs used to query the database, if None, all assay subjects are returned.

            summarized (bool): Flag indicating whether the returned assay subject records should be summarized or if full assay subject objects
                should be returned. Default is True.

        Returns:
            - A list of assay subject records (summarized or fully detailed). Depending on the summarized parameter each property in the list is listed below:

            Assay Subject record structure:
                - id (str): ID of the Assay Subject (summarized and full).
                - name (str): Name of the Assay Subject (summarized and full).
                - assaySubjectClass (dict): A JSON with assay subject class information (summarized and full).
                - descriptors (List[dict]): A list of JSON records with the assay subject descriptors information (full).
                - assaySubjectGroups (List[dict]): A list of JSON records with the assay subject groups information (full).
                - experiments (List[dict]): A list of JSON records with the assay subject experiments information (full).
                - assays (List[dict]): A list of JSON records with the assay subject assays information (full).
        """

        url = ""
        params: Dict[str, Any] = {"summarized": str(summarized).lower()}
        if isinstance(assay_subject_ids, list):
            url = self.get_assay_subjects_url
            params["ids[]"] = assay_subject_ids
        elif isinstance(assay_subject_ids, int):
            url = self.get_assay_subject_url.format(assay_subject_ids)
        elif assay_subject_ids is None:
            url = self.get_assay_subjects_url
        else:
            raise TypeError(
                f"Argument 'assay_subject_ids' must of type int or List[int]. Not type: {type(assay_subject_ids)}"
            )
        response: Dict[str, Any] = get(
            url=url,
            params=params,
            headers=self.headers,
        )

        # response["content"] = [{"id" : str, "name": str}, ...]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def delete_assay_subjects(self, assay_subject_ids: Union[int, List[int]]):
        """
        This function deletes one or many assay subject records from TEST. It receives an int ID or a list of int IDs
        through the 'assay_subject_ids' argument, which correspond to the the assay subject IDs.
        """
        params = {}
        if isinstance(assay_subject_ids, list):
            params["ids[]"] = assay_subject_ids
        elif isinstance(assay_subject_ids, int):
            params["ids[]"] = [assay_subject_ids]
        else:
            raise TypeError(
                f"Argument 'assay_subject_ids' must of type int or List[int]. Not type: {type(assay_subject_ids)}"
            )

        response = delete(
            url=self.delete_assay_subject_url.format(""),
            params=params,
            headers=self.headers,
        )

        return response["content"]

    def put_assay_subject_descriptors(
        self,
        mapper: List[dict],
        file_id: Optional[int] = None,
        filepath: Optional[str] = None,
        createSubjectsFromFile: Optional[bool] = False,
    ):
        """
            Calls Teselagen TEST API endpoint: `PUT /assay-subjects/descriptors`.
            The data can be passed via a local filepath or either the file ID after already uploading it.

            Args:
                mapper (List[dict]): This is the JSON mapper used by the endpoint to understand each of the file columns. This mapper
                    should be a list of Python Dictionary representing each structured header with a 'name', 'class' and 'subClassId' key.
                    For more information on the mappers structure refer to https://api-docs.teselagen.com/#operation/SubjectsPutAssaySubjectDecriptors
                file_id (Optional[int]) : File identifier.
                filepath (Optional[str]) : Local location of the file.
                createSubjectsFromFile (bool) : Flag that indicates whether to create new Assay Subject found in the file.

            Returns: a JSON object with a success status, the number of results inserted, and whether new assay subjects
                were created during the insert.
        """
        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None):
            if filepath is not None and (Path(filepath).exists()):
                file = self.upload_file(filepath=filepath)
                file_id = file['id']
            else:
                raise FileNotFoundError(f"File: {filepath} not found")

        body = {
            "fileId": file_id,
            "mapper": mapper,
            "createSubjectsFromFile": createSubjectsFromFile
        }

        response: Dict[str, Any] = put(
            url=self.put_assay_subject_descriptors_url,
            headers=self.headers,
            json=body,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def import_assay_subject_descriptors(
        self,
        mapper: List[dict],
        file_id: Optional[int] = None,
        filepath: Optional[str] = None,
        createSubjectsFromFile: Optional[bool] = False,
    ):
        #TODO: This is a temporary implementation while a better solution
        # for long task is implemented.

        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None):
            if filepath is not None and (Path(filepath).exists()):
                file = self.upload_file(filepath=filepath)
                file_id = file['id']
            else:
                raise FileNotFoundError(f"File: {filepath} not found")
        body = {
            "fileId": file_id,
            "mapper": mapper,
            "createSubjectsFromFile": createSubjectsFromFile
        }
        response: Dict[str, Any] = post(
            url=self.post_assay_results_import_url,
            headers=self.headers,
            json=body,
        )
        parsed_content = json.loads(response["content"])
        if 'message' in parsed_content.keys():
            parsed_content['message'] = parsed_content['message'].replace(
                "Assay results", "Assay Subject descriptor")
        response["content"] = parsed_content
        return response["content"]

    def get_assay_subjects_descriptor_import_status(
        self,
        importId: str,
    ) -> Any:
        '''
        Calls Teselagen TEST API endpoint: `GET /assays/results/import/:importId`.
        Args:
            - importId (string): ID of an assay result import process.
        Returns: a JSON object with information on the status of an assay result import job.
        '''
        try:
            response: Dict[str, Any] = get(
                url=self.get_assay_results_import_url.format(importId),
                headers=self.headers,
            )
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])
        response.pop('url')
        return response

    # Experiments Endpoints

    def get_experiments(self) -> List[Dict[str, Any]]:
        """

        Fetches all experiments from the Laboratory selected with the select_laboratory function.

        Args :

        Returns :
            () : A list of experiments objects.

        ```
            [
                {"id": "1", "name": "Experiment 1"},
                {"id": "2", "name": "Experiment 2"}
            ]
        ```

        """

        response: Dict[str, Any] = get(
            url=self.get_experiments_url,
            headers=self.headers,
        )

        # response["content"] = [{"id" : str, "name": str}, ...]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def create_experiment(self, experiment_name: str) -> Dict[str, Any]:

        experiment: List[Any] = []
        exp_response: Dict[str, Any] = {}
        experiments = self.get_experiments()
        experiment = list(
            filter(lambda x: x['name'] == experiment_name, experiments))
        if len(experiment) != 1:
            body = {"name": experiment_name}
            response: Dict[str, Any] = post(url=self.create_experiment_url,
                                            headers=self.headers,
                                            json=body)
            exp_response = json.loads(response["content"])[0]
            # Now we GET experiments from db in order to return
            # the complete experiment to the user
            experiment = list(
                filter(
                    lambda x: x['id'] == exp_response['id'],
                    self.get_experiments(),
                ))
        if len(experiment) == 0:
            raise IOError(
                f"Error while looking for new id {exp_response['id']}")

        return experiment[0]

    def delete_experiment(self, experiment_id: int) -> None:
        """ Deletes an experiment with ID=`experiment_id`. """
        response = delete(
            url=self.delete_experiment_url.format(experiment_id),
            headers=self.headers,
        )

        return None

    # Assay Endpoints

    def get_assays(
        self,
        experiment_id: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
            Fetches all assays from the experiment specified in `experiment_id`.
            If no `experiment_id` is passed, all assays from the selected
            Laboratory are returned.

            Args :

                experiment_id (int):
                    Experiment identifier.

            Returns :
                (List[Dict[str, Any]]) :
                    A list of assays objects.

            ```
                [
                    {
                        "id"         : "1",
                        "name"       : "Assay 1",
                        "experiment" : {
                                            "id"   : "1",
                                            "name" : "Experiment 1"
                                        }
                    },
                    {
                        "id"         : "2",
                        "name"       : "Assay 2",
                        "experiment" : {
                                            "id"   : "1",
                                            "name" : "Experiment 1"
                                        }
                    },
                ]
            ```
        """

        response: Dict[str, Any] = get(
            url=self.get_assays_by_experiment_url.format(experiment_id)
            if experiment_id else self.get_assays_url,
            headers=self.headers,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def create_assay(
        self,
        experiment_id: str,
        assay_name: str,
        parser_id: Optional[int] = None,
    ) -> Dict[str, Any]:

        body = {
            "name": assay_name,
            "parserId": str(parser_id) if parser_id else None
        }

        try:
            response: Dict[str, Any] = post(
                url=self.create_assay_url.format(experiment_id),
                headers=self.headers,
                json=body,
            )
        except Exception as e:
            # TODO : Use a logger
            raise Exception(e)

        assay_res = json.loads(response["content"])[0]
        # Retrieve the created object
        assay = list(
            filter(
                lambda x: x['id'] == assay_res['id'],
                self.get_assays(experiment_id=experiment_id),
            ))
        if len(assay) == 0:
            raise IOError(f"Can't find new id {assay_res['id']}")
        return assay[0]

    def delete_assay(self, assay_id: str) -> Any:
        """ Deletes an Assay with ID=`assay_id`. """
        response: Dict[str, Any] = delete(
            url=self.delete_assay_url.format(assay_id),
            headers=self.headers,
        )

        content = json.loads(response['content'])

        return content

    def delete_assays(self, assay_ids: List[str]) -> List[Any]:
        """ Deletes assays referenced by the IDs in the assay_ids list. """
        response: List[Any] = []
        for assay_id in assay_ids:
            res = self.delete_assay(assay_id=assay_id)
            response.append(res)

        return response

    def put_assay_results(
        self,
        mapper: List[dict],
        assay_id: Optional[Union[int, str]] = None,
        file_id: Optional[Union[int, str]] = None,
        filepath: Optional[Union[str, Path]] = None,
        assay_name: Optional[str] = None,
        experiment_id: Optional[Union[str, int]] = None,
        createSubjectsFromFile: Optional[bool] = True,
        createMeasurementTargetsFromFile: Optional[bool] = True,
    ):
        """
            Calls Teselagen TEST API endpoint: `PUT /assays/:assayId/results`.
            The data can be passed via a local filepath or either the file ID after already uploading it.

            Args:
                mapper (List[dict]): This is the JSON mapper used by the endpoint to understand each of the file columns. This mapper
                    should be a list of Python Dictionary representing each structured header with a 'name', 'class' and 'subClassId' key.
                    For more information on the mappers structure refer to https://api-docs.teselagen.com/#operation/AssaysPutAssayResults
                assay_id (int) : Assay identifier.
                file_id (int) : File identifier.
                filepath (int) : Local location of the file.
                assay_name (str) : Name of the assay into which insert the assay results.
                experiment_id (number) : Experiment identifier. Only used when passed and 'assay_name' an no 'assay_id'.
                createSubjectsFromFile (bool) : Flag that indicates whether to create new Assay Subject found in the file.
                createMeasurementTargetsFromFile (bool) : Flag that indicates whether to create new Measurement Target metadata records found in the file.

            Returns: a JSON object with a success status, the number of results inserted, and whether new assay subjects and/or mesurement targets
                were created during the insert.
        """
        if (assay_id is None and assay_name is not None):
            # Supports creating a new assay by providing an assay name and an experiment ID.
            assay_id = self.get_or_create_assay(
                assay_name=assay_name,
                experiment_id=str(experiment_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'assay_id' or 'assay_name'.")
        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None and filepath is not None):
            file_id = self.get_or_upload_file(
                filepath=filepath,
                assay_id=str(assay_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'file_id' or an existant 'filepath'.")

        body = {
            "assayId": assay_id,
            "fileId": file_id,
            "mapper": mapper,
            "createSubjectsFromFile": createSubjectsFromFile,
            "createMeasurementTargetsFromFile": createMeasurementTargetsFromFile
        }
        try:
            response: Dict[str, Any] = put(
                url=self.assay_results_url.format(assay_id),
                headers=self.headers,
                json=body,
            )
            print(f"response: {response}")
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])

        return response["content"]

    def import_assay_results(
        self,
        mapper: List[dict],
        assay_id: Optional[Union[int, str]] = None,
        file_id: Optional[Union[int, str]] = None,
        filepath: Optional[Union[str, Path]] = None,
        assay_name: Optional[str] = None,
        experiment_id: Optional[Union[str, int]] = None,
    ):
        """
            Calls Teselagen TEST API endpoint: `POST /assays/results/importer`.
            The data can be passed via a local filepath or either the file ID after already uploading it.

            Args:
                mapper (List[dict]): This is the JSON mapper used by the endpoint to understand each of the file columns. This mapper
                    should be a list of Python Dictionary representing each structured header with a 'name', 'class' and 'subClassId' key.
                    For more information on the mappers structure refer to https://api-docs.teselagen.com/#operation/AssaysPutAssayResults
                assay_id (int) : Assay identifier.
                file_id (int) : File identifier.
                filepath (int) : Local location of the file.
                assay_name (str) : Name of the assay into which insert the assay results.
                experiment_id (number) : Experiment identifier. Only used when passed and 'assay_name' an no 'assay_id'.

            Returns: a JSON object with a status and an import process ID. Which can be used to check the status of the import progress
                by means of the 'get_assay_results_import_status' function.
        """
        if assay_id is None and assay_name is None:
            raise Exception(
                f"Please provide a valid 'assay_id' or 'assay_name'.")

        if (assay_id is None and assay_name is not None):
            # Supports creating a new assay by providing an assay name and an experiment ID.
            assay_id = self.get_or_create_assay(
                assay_name=assay_name,
                experiment_id=str(experiment_id),
            )
        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None and filepath is not None):
            file_id = self.get_or_upload_file(
                filepath=filepath,
                assay_id=str(assay_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'file_id' or an existant 'filepath'.")
        body = {
            "assayId": assay_id,
            "fileId": file_id,
            "mapper": mapper,
        }
        try:
            response: Dict[str, Any] = post(
                url=self.post_assay_results_import_url,
                headers=self.headers,
                json=body,
            )
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])

        return response["content"]

    def get_assay_results_import_status(
        self,
        importId: str,
    ) -> Any:
        '''
        Calls Teselagen TEST API endpoint: `GET /assays/results/import/:importId`.

        Args:
            - importId (string): ID of an assay result import process.

        Returns: a JSON object with information on the status of an assay result import job.
        '''
        try:
            response: Dict[str, Any] = get(
                url=self.get_assay_results_import_url.format(importId),
                headers=self.headers,
            )
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])

        return response

    # TODO : For assays with multiple imported files, evaluate support for retrieving the assay results merged together.
    # This requires first validating that data files have a common schema or common reference column used as join key.
    def get_assay_results(
        self,
        assay_id: str,
        file_ids: Optional[List[str]] = None,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
        as_dataframe: Optional[bool] = True,
        with_subject_data: Optional[bool] = True,
        with_units: Optional[bool] = False,
    ) -> List[IAssayResults]:
        """
            Calls Teselagen TEST API endpoint: `GET /assays/:assayId/results`. It implements data pagination controllable via the
            `page_size` and `page_number` arguments further passed as query parameters to the endpoint.

            For Assays with multiple imported files, a list of dictionaries will be returned, with file information and a dataframe with its tabular content.
            You can specify which files you want to include in the results. By default a page of results from each of them will be returned.

            More information at https://api-docs.teselagen.com/#operation/AssaysGetAssayResults.

            Args:
                assay_id (str): Assay identifier.
                file_ids (Optional[List[str]]): File identifiers.
                page_size (Optional[int]): Page size for data pagination.
                page_number (Optional[int]): Page number for data pagination.
                as_dataframe (Optional[bool]): Flag indicating whether to return the data as a dataframe (default=True).
                with_subject_data (Optional[bool]): Flag indicating whether to return the assay results together with a more complete
                    information on the assay subjects (default=True).


            Returns: A IAssayResults object or a list of IAssayResults objects. The information included in the returned object is the assay results,
                plus the assay name, file information and assay subject information (if 'with_subject_data' is set to True).
        """
        if page_size is None:
            print(
                f"Using the 'page_size' argument for pagination is advised (default page_size={DEFAULT_PAGE_SIZE})."
            )

        if page_number is None:
            print(
                f"Using the 'page_number' argument for pagination is advised (default page_number=1)."
            )

        if isinstance(page_size, int) and page_size > 2 * DEFAULT_PAGE_SIZE:
            warnings.warn(
                f"Page sizes this big could end up timing out the request. Using pagination with lower page sizes is advised.",
                ResourceWarning,
            )
            print(
                f"ResourceWarning: Page sizes greater than {2 * DEFAULT_PAGE_SIZE} could end up timing out the request. Using pagination with lower page sizes is advised."
            )

        try:
            # If no file IDs are passed, query them from the Assay.
            if file_ids is None:
                # Get the files imported into the assay and only keep the ones that have successfully been imported.
                assay_files = self.get_files_info(assay_id=assay_id)
                assay_imported_files = self._filter_imported_files(assay_files)

                if len(assay_imported_files) > 0:
                    file_ids = list(
                        map(
                            lambda x: x["id"],
                            assay_imported_files,
                        ))
                else:
                    raise Exception(
                        f"Assay with ID={assay_id} has none successfully imported data files."
                    )

            final_assay_results: List[IAssayResults] = []
            for file_id in file_ids:
                final_result: IAssayResults = self._get_assay_file_results(
                    assay_id=assay_id,
                    file_id=file_id,
                    page_number=page_number,
                    page_size=page_size,
                    as_dataframe=as_dataframe,
                    with_subject_data=with_subject_data,
                    with_units=with_units,
                )
                final_assay_results.append(final_result)

            return final_assay_results

        except Exception as e:
            print(e)
            raise Exception("Error fetching assay results")

    def _get_assay_file_results(
        self: "TESTClient",
        assay_id: str,
        file_id: str,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
        as_dataframe: Optional[bool] = True,
        with_subject_data: Optional[bool] = True,
        with_units: Optional[bool] = False,
    ) -> IAssayResults:
        """
            Calls Teselagen TEST API endpoint: `GET /assays/:assayId/results?fileId=file_id`. It implements data pagination controllable via the
            `page_size` and `page_number` arguments passed as query parameters to the endpoint.

            Returns assay results from a specific imported assay file.

            More information at https://api-docs.teselagen.com/#operation/AssaysGetAssayResults.

            Args:
                assay_id (str): Assay identifier.
                file_id (str): File identifier.
                page_size (Optional[int]): Page size for data pagination.
                page_number (Optional[int]): Page number for data pagination.
                as_dataframe (Optional[bool]): Flag indicating whether to return the data as a dataframe (default=True).
                with_subject_data (Optional[bool]): Flag indicating whether to return the assay results together with a more complete
                    information on the assay subjects (default=True).


            Returns: A IAssayResults object. The information included in the returned object is the assay results,
                plus the assay name, file information and assay subject information (if 'with_subject_data' is set to True).
        """
        try:
            # NOTE: depending on the different flags, the order of the columns may vary.
            api_result = self._get_assay_file_results_from_api(
                assay_id=assay_id,
                file_id=file_id,
                page_number=page_number,
                page_size=page_size,
            )
            assay_name = api_result['name']
            assay_results = api_result['results']

            if len(assay_results) == 0:
                raise Exception(
                    f"Error getting assay results from assay with ID={assay_id}. Make sure assay has imported data files."
                )

            tabular_assay_results, assay_result_indexes = self._tabular_format_assay_result_data(
                assay_results,
                with_units,
            )

            if as_dataframe:
                final_results = pd.DataFrame(tabular_assay_results).set_index(
                    assay_result_indexes[0])
                # final_results.insert(0, "Assay", assay_name) // This column is redundant
                # If required, group by the assay results and assay subject indexes.
                # Usually these indexes are going to be the assay subject id and any reference dimension found in the assay results.
                final_results = final_results.groupby(
                    by=[*assay_result_indexes]).first().reset_index()

                if with_subject_data:
                    assaySubjectIds = list(
                        set([
                            assay_result['assaySubjectId']
                            for assay_result in assay_results
                        ]))
                    # assay_subjects = [assaySubject for assaySubject in tqdm(self.get_assay_subjects(assaySubjectIds))]
                    assay_subjects = self.get_assay_subjects(
                        assay_subject_ids=assaySubjectIds, summarized=False)

                    tabular_assay_subjects, assay_subject_indexes =\
                        self._tabular_format_assay_subject_data(assay_subjects)

                    assay_subjects_df = pd.DataFrame(
                        tabular_assay_subjects).set_index(assay_subject_indexes)

                    # Here we merge both dataframes.
                    final_results = assay_subjects_df.merge(
                        final_results,
                        left_on=assay_subject_indexes,
                        right_on=assay_subject_indexes,
                    )

            else:
                if with_subject_data:
                    assaySubjectIds = list(
                        set([
                            assay_result['assaySubjectId']
                            for assay_result in assay_results
                        ]))
                    # assay_subjects = [assaySubject for assaySubject in tqdm(self.get_assay_subjects(assaySubjectIds))]
                    assay_subjects = self.get_assay_subjects(
                        assay_subject_ids=assaySubjectIds,
                        summarized=False,
                    )
                    tabular_assay_subjects, assay_subject_indexes =\
                        self._tabular_format_assay_subject_data(assay_subjects)

                    final_results = [{
                        **{
                            "Assay": assay_name
                        },
                        **assay_subject,
                        **assay_result
                    } for (assay_subject, assay_result) in zip(
                        tabular_assay_subjects, tabular_assay_results)]
                else:
                    final_results = [{
                        **{
                            "Assay": assay_name
                        },
                        **assay_result
                    } for assay_result in tabular_assay_results]

            assayResults: IAssayResults = {
                "assayId": assay_id,
                "fileId": file_id,
                "data": final_results
            }

            return assayResults

        except Exception as e:
            raise Exception("Error fetching assay file results")

    def _get_assay_file_results_from_api(
        self,
        assay_id: str,
        file_id: str,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Any:

        url = self.assay_results_url.format(assay_id)

        params = {
            "fileId":
                file_id,
            "pageNumber":
                page_number if page_number is not None else 1,
            "pageSize":
                page_size if page_size is not None else DEFAULT_PAGE_SIZE
        }

        api_result = None
        try:
            response: Dict[str, Any] = get(
                url=url,
                headers=self.headers,
                params=params,
            )
            api_result = json.loads(response["content"])

        except Exception as e:
            print(e)
            raise Exception(
                f"Error getting assay results from assay with ID={assay_id}. Make sure assay exists or that has imported results."
            )

        return api_result

    # File Endpoints

    def get_files_info(
        self,
        experiment_id: Optional[str] = None,
        assay_id: Optional[str] = None,
        file_id: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
            Fetches all files from the selected Laboratory. You can also filter the results by experiment, by assay or by file, using the
            'experiment_id', 'assay_id' and/or 'file_id' arguments.

            Returns :
                () :
                    A list of assays objects.

            ```
                [{
                    "id": "1",
                    "name": "File 1",
                    "assay": {
                        "id": "1",
                        "name": "Assay 1"
                    },
                    "experiment": {...}
                },
                {
                    "id": "2",
                    "name": "File 2",
                    "assay": null,
                    "experiment": {...}
                }]
            ```
        """

        params = {"experimentId": experiment_id, "assayId": assay_id}

        # Removes params with None values.
        params = {
            key: value for key, value in params.items() if value is not None
        }

        response: Dict[str, Any] = get(
            url=self.get_files_info_url,
            headers=self.headers,
            params=params,
        )

        results: List[Dict[str, Any]] = json.loads(response["content"])


        if file_id is not None:
            results = list(filter(lambda x: x['id'] == file_id, results))

        return results

    def upload_file(
        self,
        filepath: Union[str, Path],
        experiment_id: Optional[int] = None,
        assay_id: Optional[str] = None,
    ):
        """
            Uploads a file. The request body is of type "multipart/form-data".

            It requires the "filepath" and optionally with an "assay_id".

            If no `assay_id` is passed the file will be uploaded linked to no
            assay.

            NB: If an assay_id with an assigned parser is passed the file will be
            automatically parsed with such parser.

            Args :
                filepath (str) :
                    Path to the file to be uploaded.

                experiment_id (Optional[int]) :
                    Experiment identifier.

                assay_id (Optional[int]) :
                    Assay identifier.

            Returns :
        """

        filepath = Path(filepath)
        multipart_form_data = {
            'file': (filepath.name, open(filepath, 'rb')),
        }

        # We need a header file wihtout the 'Content-Type' key because this is a 'multipart/form-data' request
        # unlike the others which have Content-Type = 'application/json'. Here, only the authorization token is needed.
        headers = self.headers.copy()
        del headers['Content-Type']

        upload_file_url = self.upload_file_into_assay_url.format(
            assay_id
        ) if assay_id else self.upload_file_into_experiment_url.format(
            experiment_id) if experiment_id else self.upload_file_url
        response: Dict[str, Any] = post(
            url=upload_file_url,
            headers=headers,
            files=multipart_form_data,
        )
        res_files_info = json.loads(response["content"])
        if not isinstance(res_files_info, dict):
            raise IOError(
                f"There was a problem with upload (maybe check assay_id): response: {response}"
            )
        # Build our object to be returned (new file_info)
        # Get the file info with the right id
        files_info = list(
            filter(
                lambda x: x['id'] == res_files_info['id'],
                self.get_files_info(),
            ))
        if len(files_info) == 0:
            raise IOError(
                f"Name {multipart_form_data['file'][0]} not found in uploaded files"
            )
        return files_info[0]

    def download_file(self, file_id: Union[int, str]) -> StringIO:
        """
            It will return the data contents of the corresponding file with ID
            specified in the "file_id" argument.

            It returns a StringIO object, which keeps the file data serialized.
            One could take this serialized data and write it into a file or directly rad it with pandas.

            Args:
                file_id (int) :
                        File identifier.

            Returns:
                a StringIO object with the data.
        """

        response: Dict[str, Any] = get(
            url=self.get_file_data_url.format(file_id),
            headers=self.headers,
        )

        return StringIO(response["content"])

    def delete_file(self, file_id: Union[int, str]) -> Dict[str, Any]:
        """ Deletes a File with ID=`file_id`. """
        response: Dict[str, Any] = delete(
            url=self.delete_file_url.format(file_id),
            headers=self.headers,
        )

        content = json.loads(response["content"])

        return content

    # Metadata Endpoints

    def get_metadata(self, metadataType: str, metadataTypeFields: str = None):
        """

        Returns metadata records according to the 'metaDataType' path parameter. Available metaDataTypes are:

            - assaySubjectClass
            - measurementTarget
            - measurementType
            - unit
            - unitScale
            - unitDimension
            - descriptorType

        Args :

            metadataType (str): The type of a metadata. Must be one of the available metadata types listed above.

        Returns : A JSON object with the metadata records belonging to the requested metadata type.

        ```
            [
                {"id": "1", "name": "Metadata Record 1"},
                {"id": "2", "name": "Metadata Record 2"},
            ]
        ```

        """

        response: Dict[str, Any] = get(
            url=self.get_metadata_url.format(metadataType),
            headers=self.headers,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def create_metadata(
        self,
        metadataType: str,
        metadataRecord: Union[List[dict], dict],
    ):
        """
            Calls Teselagen TEST API endpoint: `POST /metadata`.
            More information at https://api-docs.teselagen.com/#operation/MetadataCreateMetadata.

            Args:
                metadataType (str): Name of the metadata type/class.
                metadataRecord (Union[List[dict], dict]): Either an array of metadata records or a single one.
                    These should follow the required structure of a metadata record. For more information on this
                    refer to the above API documentation link.
        """
        body = {"metaData": {metadataType: metadataRecord}}

        response: Dict[str, Any] = post(url=self.create_metadata_url,
                                        headers=self.headers,
                                        json=body)

        # [{ id: "3" }]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def delete_metadata(self, metadataType: str, metadataId: int):
        response = delete(
            url=self.delete_metadata_url.format(
                metadataType,
                metadataId,
            ),
            headers=self.headers,
        )

        # response["content"] = json.loads(response["content"])

        return True

    # TEST Client Utils

    def get_or_create_assay(
        self,
        assay_name: str,
        experiment_id: str,
    ) -> str:
        ''' Supports creating a new assay by providing an assay name and an experiment ID.'''
        assay_id = None
        if (experiment_id is not None):
            assays = self.get_assays()
            assay = list(
                filter(
                    lambda x: x['name'] == assay_name and x['experiment']['id']
                    == experiment_id, assays))
            assay_id = assay[0]['id'] if len(assay) > 0 else self.create_assay(
                experiment_id=experiment_id, assay_name=assay_name)['id']
        else:
            raise Exception(f"Please provide a valid 'experiment_id'.")

        return assay_id

    def get_or_upload_file(
        self,
        filepath: Union[str, Path],
        assay_id: str,
    ) -> Optional[str]:
        file_id = None
        if Path(filepath).exists():
            # See the current files already uploaded to the assay.
            files = self.get_files_info()
            assay_files = list(
                filter(
                    lambda x: x['assay'] is not None and x['assay']['id'] ==
                    assay_id and filepath is not None and Path(x[
                        'name']).name == Path(filepath).name,
                    files,
                ))

            if (len(assay_files) > 0):
                file_id = assay_files[0]['id']
            # NOTE: When a file with the same name has already been uploaded into the Assay, do not upload the file again.
            else:
                file = self.upload_file(
                    filepath=filepath,
                    assay_id=assay_id,
                )
                file_id = file['id']
        else:
            raise FileNotFoundError(f"Provided 'filepath' does not exist.")
        return file_id

    def _tabular_format_assay_subject_data(self, assay_subjects_data: Any):
        tabular_assay_subjects = []
        for assay_subject_data in assay_subjects_data:
            assay_subject_row_dict = {
                "Subject ID": assay_subject_data['id'],
                "Subject Name": assay_subject_data['name'],
                "Subject Class": assay_subject_data['assaySubjectClass']['name']
            }
            for descriptor in assay_subject_data['descriptors']:
                assay_subject_row_dict[descriptor['descriptorType']
                                       ['name']] = descriptor['value']

            tabular_assay_subjects.append(assay_subject_row_dict)

        indexes = ["Subject ID"]
        return tabular_assay_subjects, indexes

    def _tabular_format_assay_result_data(
        self,
        assay_result_data: Any,
        with_units: Optional[bool] = False,
    ):
        tabular_assay_results = []
        assaySubjectColumnName = 'Subject ID'
        assaySubjectIds = set()
        referenceDimensions = set()
        measurementTypes = set()
        for result in assay_result_data:
            # The assay subject ID is important because a tabular form would be indexed by these.
            assaySubjectId = result['assaySubjectId']
            assaySubjectIds.add(assaySubjectId)
            tabular_row_assay_result_dict = {
                assaySubjectColumnName: assaySubjectId
            }

            # reference dimensions are important when formatting assay results, because a tabular form
            # would be indexed by these.
            if "reference" in result:
                referenceDimension = result['reference']['name']
                referenceDimensions.add(referenceDimension)
                tabular_row_assay_result_dict[referenceDimension] = result[
                    'reference']['value']

                # Here if 'with_units' is True (False by default) unit columns will be returned.
                if with_units:
                    tabular_row_assay_result_dict[
                        f"{referenceDimension} Metric"] = result['reference'][
                            'unit']

            measurementType = result['result']['name']
            measurementTypes.add(measurementType)
            tabular_row_assay_result_dict[measurementType] = result['result'][
                'value']

            # Here if 'with_units' is True (False by default) unit columns will be returned.
            if with_units:
                tabular_row_assay_result_dict[
                    f"{measurementType} Metric"] = result['result']['unit']

            tabular_assay_results.append(tabular_row_assay_result_dict)

        if len(referenceDimensions) > 1:
            # TODO: add support for multiple reference dimensions.
            raise Exception("Multiple Reference Dimensions not supported.")

        indexes = [assaySubjectColumnName, *list(referenceDimensions)]

        return tabular_assay_results, indexes

    @staticmethod
    def _filter_imported_files(files: List[Any]) -> List[Any]:
        """ Takes in a list of file IDs and returns only those that are successfully imported in TEST """
        if isinstance(files, list) and len(files) > 0:
            filtered_files = list(
                filter(
                    lambda x: "id" in x and "importStatus" in x and x[
                        "importStatus"] in IMPORTED_FILE_STATUSES, files))
            return filtered_files
        else:
            return []



#!/usr/bin/env python3
from typing import List, Dict, Any, Optional, Tuple, Union
import uuid
from IPython.display import display_javascript, display_html, display
import json

from dna_features_viewer import GraphicFeature, CircularGraphicRecord #, GraphicRecord
from SecretColors import Palette
from matplotlib.axes import SubplotBase


def plot_plasmid_features(
    plasmid_length: int, 
    features: List[Dict[str, Any]], 
    figure_width: int=5, 
    palette: Optional[Palette]=None)->Tuple[SubplotBase, Tuple[Any, Any]]:
    """Plots features in a circular dna sequence

    Args:
        plasmid_length (int): Number of nucleotid bases of the plasmid sequence
        features (List[Dict[str, Any]]): Features as obtained from TeselaGen DNA Sequence object
        figure_width (int, optional): Width size of figure. Defaults to 5.
        palette (Optional[Palette], optional): A SecretColors color palette. Defaults to None, meaning 
            `Palette("material")` will be used.

    Returns:
        Tuple[AxesSubplot, Tuple[Any, Any]]: Axes and a tuple with Graphic features data
    """
    # Define random color pallete
    if palette is None:
        palette = Palette("material")
    colors = palette.random(no_of_colors=len(features))
    # Create feat objects
    plot_feats = [
        GraphicFeature(start=feat['start'], end=feat['end'], strand=1*feat['forward'], label=feat['name'], color=colors[i]) 
        for i, feat in enumerate(features)]
    # Make graphic record and plot
    record = CircularGraphicRecord(sequence_length=plasmid_length, features=plot_feats)
    ax, _ = record.plot(figure_width=figure_width)
    return record.plot(ax)



class RenderJSON(object):
    """ Provides a an interactive visualization for json (or serializable list/dict) objects"""
    def __init__(
        self, 
        json_data: Union[dict, list, str],
        height: str="max-content",
        width: str="100%",
        background_color: str= "#f2f3ff"
        ):
        """Initiates json interactive visualization object for IPython

        Args:
            json_data (Union[dict, list, str]): The object to be shown.

        Raises:
            TypeError: If input type is not supported
        """
        if isinstance(json_data, (dict, list)) :
            self.json_str = json.dumps(json_data)
        elif isinstance(json_data, str):
            self.json_str = json_data
        else:
            raise TypeError(f"Can't process json_data of type {type(json_data)}")
        self.style={"height": height, "width": width, "background_color": background_color}
        self.uuid = str(uuid.uuid4())

    def _ipython_display_(self):
        """Renders JSON into collapsible HTML object by using [renderjson.js](https://github.com/caldwell/renderjson)
        """
        display_html(f"""
            <div id="{self.uuid}" style="height: {self.style["height"]}; width:{self.style['width']};background-color: {self.style["background_color"]}";></div>
            """, raw=True)
        display_javascript("""
            require(["https://rawgit.com/caldwell/renderjson/master/renderjson.js"], function() {
            document.getElementById('%s').appendChild(renderjson(%s))
            });
            """ % (self.uuid, self.json_str), raw=True)